from abc import ABC, abstractmethod
from datetime import date, datetime
from typing import Optional
from ...lstindex import LSTInterval


class LocationProvider(ABC):
    @staticmethod
    @abstractmethod
    def calc_sun(date: date, latitude: Optional[float], longitude: Optional[float]) -> dict:
        pass

    @staticmethod
    @abstractmethod
    def calc_intervals(
        latitude: str | float, longitude: str | float, today_dt: datetime, self
    ) -> list[LSTInterval]:
        pass



from ..lstcalendar import LSTCalendar
from .Observation import Observation
from typing import Optional
from perf import track_total_runtime
from conf import Conf, LocationProviderType

conf = Conf()


@track_total_runtime
def is_observable(
    observation: Observation,
    yyyymmdd_start: Optional[str] = None,
    yyyymmdd_end: Optional[str] = None,
    lstCalendar: Optional[LSTCalendar] = None,
    latitude: Optional[str | float] = conf.LATITUDE,
    longitude: Optional[str | float] = conf.LONGITUDE,
    provider: Optional[LocationProviderType] = None,
) -> bool:
    """
    Determines if an Observation is observable within the specified date and location parameters.

    Parameters
    ----------
    observation : Observation
        The Observation object to be checked.
    yyyymmdd_start : str, optional
        (Optional) The start date in the format 'YYYYMMDD'.
    yyyymmdd_end : str, optional
        (Optional) The end date in the format 'YYYYMMDD'. If not provided, the start date is used.
    lstCalendar : LSTCalendar, optional
        (Optional) An instance of LSTCalendar. If not provided, a new LSTCalendar instance will be created based on the date and location parameters.
    latitude : str or float, optional
        (Optional) The latitude for the observation in the format 'D:M:S'. Defaults to -30:42:39.8 (for ASTRAL provider).
    longitude : str or float, optional
        (Optional) The longitude for the observation in the format 'D:M:S'. Defaults to 21:26:38.0 (for ASTRAL provider).

    Returns
    -------
    bool
        True if the observation is observable within the specified parameters, False otherwise.
    """

    # TODO fix me. Some import problem
    from ..lstcalendar import LSTCalendar

    yyyymmdd_end = yyyymmdd_end if yyyymmdd_end else yyyymmdd_start

    # Create an LSTCalendar instance if not provided
    calendar = (
        lstCalendar
        if lstCalendar
        else LSTCalendar(
            yyyymmdd_start, yyyymmdd_end, latitude=latitude, longitude=longitude, provider=provider
        )
    )

    return bool(observation.observables(calendar))



from .AstroUtils import AstroUtil
import numpy as np
import math


class SolarSystem:
    @staticmethod
    def helios(p, jd):
        T = (jd - 2451545.0) / 36525
        T2 = T * T
        T3 = T2 * T
        N = AstroUtil.rev360(p["N"][0] + p["N"][1] * T + p["N"][2] * T2 + p["N"][3] * T3)
        i = p["i"][0] + p["i"][1] * T + p["i"][2] * T2 + p["i"][3] * T3
        L = AstroUtil.rev360(p["L"][0] + p["L"][1] * T + p["L"][2] * T2 + p["L"][3] * T3)
        a = p["a"][0] + p["a"][1] * T + p["a"][2] * T2 + p["a"][3] * T3
        e = p["e"][0] + p["e"][1] * T + p["e"][2] * T2 + p["e"][3] * T3
        P = AstroUtil.rev360(p["P"][0] + p["P"][1] * T + p["P"][2] * T2 + p["P"][3] * T3)
        M = AstroUtil.rev360(L - P)
        w = AstroUtil.rev360(L - N - M)

        E0 = M + (AstroUtil.RTD * e * AstroUtil.sind(M) * (1 + e * AstroUtil.cosd(M)))
        E = E0 - (E0 - AstroUtil.RTD * e * AstroUtil.sind(E0) - M) / (1 - e * AstroUtil.cosd(E0))

        while abs(E0 - E) > 0.0005:
            E0 = E
            E = E0 - (E0 - AstroUtil.RTD * e * AstroUtil.sind(E0) - M) / (
                1 - e * AstroUtil.cosd(E0)
            )

        x = a * (AstroUtil.cosd(E) - e)
        y = a * math.sqrt(1 - e * e) * AstroUtil.sind(E)
        r = math.sqrt(x * x + y * y)
        v = AstroUtil.rev360(AstroUtil.atan2d(y, x))

        xeclip = r * (
            AstroUtil.cosd(N) * AstroUtil.cosd(v + w)
            - AstroUtil.sind(N) * AstroUtil.sind(v + w) * AstroUtil.cosd(i)
        )
        yeclip = r * (
            AstroUtil.sind(N) * AstroUtil.cosd(v + w)
            + AstroUtil.cosd(N) * AstroUtil.sind(v + w) * AstroUtil.cosd(i)
        )
        zeclip = r * AstroUtil.sind(v + w) * AstroUtil.sind(i)

        return [xeclip, yeclip, zeclip]

    @staticmethod
    def radecr(obj, base, jd):
        # Equatorial co-ordinates
        x = obj[0]
        y = obj[1]
        z = obj[2]

        # Obliquity of Ecliptic
        obl = 23.4393 - 3.563e-7 * (jd - 2451543.5)

        # Convert to Geocentric co-ordinates
        x1 = x - base[0]
        y1 = (y - base[1]) * AstroUtil.cosd(obl) - (z - base[2]) * AstroUtil.sind(obl)
        z1 = (y - base[1]) * AstroUtil.sind(obl) + (z - base[2]) * AstroUtil.cosd(obl)

        # RA and dec
        ra = AstroUtil.rev360(AstroUtil.atan2d(y1, x1)) / 15.0
        dec = AstroUtil.atan2d(z1, math.sqrt(x1 * x1 + y1 * y1))

        # Earth distance
        r = math.sqrt(x1 * x1 + y1 * y1 + z1 * z1)

        return [ra, dec, r]

    @staticmethod
    def equ_to_horizon(obj, jd):
        from .DateTimeUtils import DateTimeUtil

        ra = obj["ra"]
        dec = obj["dec"]

        # Calculate Hour Angle
        H = DateTimeUtil.local_sidereal(jd) - ra
        sinH = AstroUtil.sind(H)
        cosH = AstroUtil.cosd(H)

        sinl = AstroUtil.sind(AstroUtil.MKAT_POSITION["longitude"])
        cosl = AstroUtil.cosd(AstroUtil.MKAT_POSITION["latitude"])
        sind = AstroUtil.sind(dec)
        cosd = AstroUtil.cosd(dec)
        tand = AstroUtil.tand(dec)

        # Calculate Azimuth and Elevation
        az = AstroUtil.rev360(AstroUtil.atan2d(sinH, cosH * sinl - tand * cosl) + 180)
        el = AstroUtil.asind(sinl * sind + cosl * cosd * cosH)

        obj["azimuth"] = az
        obj["elevation"] = el

        return [az, el]

    @staticmethod
    def g_sidereal(jd):
        T = (jd - 2451545.0) / 36525
        res = (
            280.46061837
            + 360.98564736629 * (jd - 2451545.0)
            + 0.000387933 * T**2
            - T**3 / 38710000
        )
        return AstroUtil.rev360(res)

    @staticmethod
    def sunpos(planet, jd):
        sun_xyz = np.array([0.0, 0.0, 0.0])
        planet_xyz = SolarSystem.helios(planet, jd)
        radec = SolarSystem.radecr(sun_xyz, planet_xyz, jd)
        ra = 15 * radec[0]
        dec = radec[1]
        altaz = SolarSystem.equ_to_horizon({"ra": ra, "dec": dec}, jd)

        return {
            "ra": ra,
            "dec": dec,
            "azimuth": altaz[0],
            "elevation": altaz[1],
            "isVisible": altaz[1] > 0,
        }



from typing import Self
from datetime import datetime, timedelta
import re
from conf import Conf, LocationProviderType
from lstpressure import LSTIntervalType as I, Observation, LST
from ..AppInterface import AppInterface
from ..aggregate import parse_sql_query, execute_csvsql, quoted_existing_reports
from logger import error
import sys
import pandas as pd
from io import StringIO

conf = Conf()

filter_mapping = {
    I.NIGHT.name: I.NIGHT,
    I.SUNRISE_SUNSET.name: I.SUNRISE_SUNSET,
    I.ALL_DAY.name: I.ALL_DAY,
    I.SUNSET_SUNRISE.name: I.SUNSET_SUNRISE,
    I.OBSERVATION_WINDOW.name: I.OBSERVATION_WINDOW,
}


def kebab_to_snake(input_string):
    return re.sub(r"-", "_", input_string)


def parseDateInput(input: str, current_date=None) -> str:
    """
    Handle relative date inputs
    """

    # Check if the input is a direct date
    if re.match(r"^\d{8}$", input):
        return input

    # Convert provided current_date from yyyymmdd format to datetime
    if current_date is not None:
        try:
            current_date = datetime.strptime(current_date, "%Y%m%d")
        except ValueError:
            raise ValueError("current_date must be in yyyymmdd format")
    else:
        current_date = datetime.now()
    input = input.upper()
    match = re.match(r"^([-+])(\d+)([DMY])$", input)
    if match:
        sign, value, unit = match.groups()
        value = int(value)

        if unit == "D":
            # Add or subtract days
            if sign == "+":
                new_date = current_date + timedelta(days=value)
            else:
                new_date = current_date - timedelta(days=value)
        elif unit == "M":
            # Add or subtract months
            if sign == "+":
                new_month = current_date.month - 1 + value
            else:
                new_month = current_date.month - 1 - value

            year = current_date.year + new_month // 12
            month = new_month % 12 + 1
            day = min(
                current_date.day,
                [
                    31,
                    29 if year % 4 == 0 and not year % 100 == 0 or year % 400 == 0 else 28,
                    31,
                    30,
                    31,
                    30,
                    31,
                    31,
                    30,
                    31,
                    30,
                    31,
                ][new_month % 12],
            )
            new_date = datetime(year, month, day)
        elif unit == "Y":
            # Add or subtract years
            if sign == "+":
                new_date = current_date.replace(year=current_date.year + value)
            else:
                new_date = current_date.replace(year=current_date.year - value)

        return new_date.strftime("%Y%m%d")

    # If the input format is unrecognized
    raise ValueError("Invalid input format")


class Observables(AppInterface):
    id = "observables"
    usage = "lstpressure observables -h"
    description = "Generate a list of observables (slots where OPT observations can be observed)"

    def __init__(self, parser) -> None:
        super().__init__(parser)

    def build(self) -> Self:
        self.parser.add_argument(
            "--start",
            type=str,
            required=False,
            default=datetime.today().strftime("%Y%m%d"),
            help="The start date in the format 'YYYYMMDD', or as a relative date (from today) in the form +|-iD|M|Y (example: \"+3D\"). Defaults to today",
            metavar="",
        )

        self.parser.add_argument(
            "--end",
            type=str,
            required=False,
            default=None,
            help="The end date in the format 'YYYYMMDD', or as a relative date (from --start) in the form +|-iD|M|Y (example: \"+1Y\"). Defaults to --start",
            metavar="",
        )

        self.parser.add_argument(
            "--input",
            required=False,
            type=str,
            help="Path to an OPT csv download. If this is omitted, provide the CSV file contents via stdin",
            metavar="",
        )

        self.parser.add_argument(
            "--output",
            type=str,
            help="Path to the output csv file",
            metavar="",
        )

        self.parser.add_argument(
            "--loc-provider",
            type=lambda s: s.upper(),
            choices=list(LocationProviderType.__members__),
            default="MEERKAT",
            help="Specify the location provider for calculating sun statistics and for calculating intervals. Choose between: ASTRAL or MEERKAT (default). MeerKAT: intervals calculated using sunrise/sunset values (including NIGHT), ASTRAL: intervals calculated using dawn, sunrise, sunset, dusk",
            metavar="",
        )

        self.parser.add_argument(
            "--filter",
            type=lambda s: kebab_to_snake(s.upper()),
            choices=list(I.__members__),
            default=None,
            required=False,
            metavar="",
            help=f"Select from: {', '.join(list(I.__members__))}",
        )

        self.parser.add_argument(
            "--lat",
            metavar="D:M:S",
            default=None,
            nargs="?",
            type=str,
            help="The latitude for the observation in the format 'D:M:S'. Default is '-30:42:39.8' (for ASTRAL provider). Value must be quoted --lat=\"location\"",
        )

        self.parser.add_argument(
            "--long",
            metavar="D:M:S",
            default=None,
            nargs="?",
            type=str,
            help="The longitude for the observation in the format 'D:M:S'. Default is '21:26:38.0' (for ASTRAL provider). Value must be quoted --long=\"location\"",
        )

        self.parser.add_argument(
            "--aggregate",
            required=False,
            type=str,
            metavar="",
            help=f'Apply a SQL aggregation to the CSV before printing to a file/stdout, or apply a built-in aggregation: {quoted_existing_reports}. The CSV is provided as a relation called "stdin"',
        )

        return self

    def parse(self, args) -> Self:
        if args.lat or args.long:
            if LocationProviderType[args.loc_provider] == LocationProviderType.MEERKAT:
                error(
                    "The MEERKAT provider has set lat/long coordinates that can't be overriden. Either use a different loc-provider or don't specify --lat and --long"
                )
                exit(1)
        if args.loc_provider == "MEERKAT":
            conf.LATITUDE = None
            conf.LONGITUDE = None
        else:
            if args.lat and args.lat != "":
                conf.LATITUDE = args.lat
            if args.long and args.long != "":
                conf.LONGITUDE = args.long
        self.input = args.input
        self.start = parseDateInput(args.start)
        self.end = parseDateInput(args.end, self.start) if args.end else self.start
        conf.LOC_PROVIDER = LocationProviderType[args.loc_provider]

        filter_name = args.filter
        if filter_name and not filter_mapping.get(filter_name.upper()):
            error(f"Invalid filter name, valid options: {', '.join(list(filter_mapping.keys()))}")
            exit(1)
        self.filter_value = (
            None if not filter_name else filter_mapping.get(filter_name.upper(), None)
        )

        self.output = args.output if args.output else None
        self.aggregate = args.aggregate if args.aggregate else None
        return self

    def exe(self) -> None:
        sql = parse_sql_query(self.aggregate) if self.aggregate else None

        def observation_filter(observation: Observation):
            if self.filter_value in observation.utc_constraints:
                return True
            return False

        if not self.input:
            if not sys.stdin.isatty():
                input_data = sys.stdin.read()
                input = pd.read_csv(StringIO(input_data))
            else:
                error(
                    "No input provided. Please either provide a filepath (--input), or pipe a CSV-formatted string from stdin"
                )
                exit(1)
        else:
            input = self.input

        output_csv_string = LST(
            input,
            calendar_start=self.start,
            calendar_end=self.end,
            observation_filter=observation_filter if self.filter_value else None,
            latitude=conf.LATITUDE,
            longitude=conf.LONGITUDE,
        ).to_csv_string()

        output_string = (
            execute_csvsql(output_csv_string, sql, self.aggregate) if sql else output_csv_string
        )

        if self.output:
            with open(self.output, "w") as f:
                f.write(output_string)
        else:
            print(output_string)

# Automatically added by katversion
__version__ = '1.3.0'



from ...observation import Observation
from ...lstindex import LSTIntervalType as I
import math
from typing import Tuple
from perf import track_total_runtime


def convert_to_hours(x):
    try:
        result = round((int(x.split(":")[0]) * 3600 + int(x.split(":")[1]) * 60) / 3600, 2)
        return result
    except:
        return None


@track_total_runtime
def calculate_constraints(row) -> list[I]:
    constraints = []
    if row["night_obs"] == "Yes":
        constraints.append(I.NIGHT)
    elif row["avoid_sunrise_sunset"] == "Yes":
        constraints.append(I.SUNRISE_SUNSET)
        constraints.append(I.SUNSET_SUNRISE)

    # If neither night_obs nor avoid_sunrise_sunset was selected
    if not constraints:
        constraints.append(I.ALL_DAY)

    return constraints


@track_total_runtime
def calculate_observations(dataFrame, observation_filter) -> Tuple[Observation]:
    try:
        # Convert durations to hours and round off
        dataFrame["duration_hours"] = dataFrame["simulated_duration"].apply(
            lambda x: round(0 if math.isnan(x) else x / 3600, 2)
        )

        # Convert LST start and end times to floating point hours
        dataFrame["lst_start_hours"] = dataFrame["lst_start"].apply(convert_to_hours)
        dataFrame["lst_end_hours"] = dataFrame["lst_start_end"].apply(convert_to_hours)

        # Build observations using list comprehension for efficiency
        observations = [
            Observation(
                id=row["id"],
                lst_window_start=row["lst_start_hours"],
                lst_window_end=row["lst_end_hours"],
                utc_constraints=calculate_constraints(row),
                duration=row["duration_hours"],
                proposal_id=row["proposal_id"],
            )
            for _, row in dataFrame.iterrows()
            if row["lst_end_hours"] - row["lst_start_hours"] > 0
        ]

        return tuple(filter(observation_filter, observations))
    except Exception as e:
        raise ValueError("Error parsing CSV - please check input.", e)



from ..AppInterface import AppInterface
from typing import Self
import os
import subprocess
from logger import error
import sys


SQL_FOLDER_PATH = os.path.join(os.path.dirname(__file__), "sql")
existing_reports = [
    f.replace(".sql", "")
    for f in os.listdir(SQL_FOLDER_PATH)
    if os.path.isfile(os.path.join(SQL_FOLDER_PATH, f))
]
quoted_existing_reports = ", ".join([f'"{r}"' for r in existing_reports])


class Aggregate(AppInterface):
    id = "aggregate"
    usage = "lstpressure aggregate -h"
    description = "Aggregate observables using built-in reports or bespoke SQL commands"

    def __init__(self, parser) -> None:
        super().__init__(parser)

    def build(self) -> Self:
        self.parser.add_argument(
            "--input",
            required=False,
            type=str,
            metavar="",
            help="Path to lst-pressure output csv file (if not provided, stdin is used instead)",
        )
        self.parser.add_argument(
            "--query",
            required=True,
            type=str,
            metavar="",
            help=f'SQL string to apply to input CSV (i.e. "select * from stdin"), or apply a built-in aggregation: {quoted_existing_reports}. By default the SQLite engine is used, and you can enable the DuckDB engine by including "duckdb" either in a filename or as a comment somewhere in the query text',
        )
        self.parser.add_argument(
            "--echo",
            required=False,
            default=False,
            action="store_true",
            help=f"Echo the aggregation query back - useful for evaluating .sql file contents that would otherwise be opaque. For example: --echo --query {existing_reports[0]}",
        )

        return self

    def parse(self, args) -> Self:
        self.args = args
        self.echo = args.echo
        self.input = args.input
        self.query = parse_sql_query(args.query)
        return self

    def exe(self) -> None:
        if self.echo:
            print(self.query)
            exit()

        if not self.input:
            if not sys.stdin.isatty():
                input_csv = sys.stdin.read()
            else:
                error(
                    "No input provided. Please either provide a filepath (--input), or pipe a CSV-formatted string from stdin"
                )
                exit(1)
        else:
            with open(self.input, "r") as file:
                input_csv = file.read()

        result = execute_csvsql(input_csv, self.query, self.args.query)

        # Print result to stdout
        print(result)
        exit()


def parse_sql_query(str_input):
    # A .sql file
    if os.path.isfile(str_input):
        with open(str_input, "r") as file:
            return file.read().strip()

    # Try to find the SQL file if it's included in the tool
    path = os.path.join(SQL_FOLDER_PATH, f"{str_input}")

    # Name of view without .sql
    if os.path.isfile(path):
        with open(path, "r") as file:
            return file.read().strip()

    # Name of view with .sql
    if os.path.isfile(path + ".sql"):
        with open(path + ".sql", "r") as file:
            return file.read().strip()

    # If not a file, assume it's a SQL string
    return str_input


def execute_csvsql(input_content, sql, query_name=""):
    duck_opts = (
        [
            "--db",
            "duckdb:///:memory:",
            "--insert",
        ]
        if "duckdb" in query_name or "duckdb" in sql
        else []
    )

    command = ["csvsql", "--query", sql, *duck_opts]

    try:
        # Use subprocess.Popen to run the command
        process = subprocess.Popen(
            command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, text=True
        )

        # Pass input content to the subprocess
        output, _ = process.communicate(input=input_content)

        # Wait for the process to finish and get the return code
        return_code = process.wait()

        if return_code == 0:
            return output
        else:
            error(f"csvsql process returned a non-zero exit code: {return_code}")
            error(f"Command output: {output}")
            return None
    except Exception as e:
        raise e


__all__ = ["parse_sql_query", "Aggregate", "execute_csvsql"]

# Automatically added by katversion
__version__ = '1.3.0'



"""
observable.Observable
"""
from datetime import datetime
from typing import Any, Self
from ..observation import Observation
from ..lstindex import LSTInterval, LSTIntervalType
from perf import track_total_runtime, decorate_all


@decorate_all(track_total_runtime)
class Observable:
    """
    Represents a valid scheduling time frame based on observational constraints.

    Attributes:
        interval (LSTInterval): The time interval or range representing the observation window.
        observation (Observation): The observational data or constraints defining the validity of the interval.

    Args:
        interval (LSTInterval): The time interval or range for the observation window.
        observation (Observation): The observational constraints or data associated with the interval.
    """

    def __init__(self, interval: LSTInterval, observation: Observation):
        """
        Initializes an instance of the Observable class.

        Args:
            interval (LSTInterval): The time interval or range for the observation window.
            observation (Observation): The observational constraints or data associated with the interval.
        """
        self._interval = interval
        self._observation = observation

    @property
    def interval(self) -> LSTInterval:
        """
        Returns the time interval or range representing the observation window.

        Returns:
            LSTInterval: The time interval or range for the observation.
        """
        return self._interval

    @property
    def observation(self) -> Observation:
        """
        Returns the observational data or constraints defining the validity of the interval.

        Returns:
            Observation: The observational constraints or data.
        """
        return self._observation

    @property
    def id(self) -> Any:
        """
        Returns the identification of the observation.

        Returns:
            Any: The identification of the observation.
        """
        return self.observation.id

    @property
    def proposal_id(self) -> Any:
        """
        Returns the proposal identification associated with the observation.

        Returns:
            Any: The proposal identification.
        """
        return self.observation.proposal_id

    @property
    def dt(self) -> datetime:
        """
        Returns the date and time of the observation.

        Returns:
            datetime: The date and time of the observation.
        """
        return self.interval.dt

    @property
    def duration(self) -> float:
        """
        Returns the duration of the observation.

        Returns:
            float: The duration of the observation in hours.
        """
        return self.observation.duration

    @property
    def utc_constraint(self) -> LSTIntervalType:
        """
        Returns the UTC constraint type of the observation interval.

        Returns:
            LSTIntervalType: The UTC constraint type of the interval.
        """
        return self.interval.type.name

    def to_tuple(self):
        """
        Converts the Observable instance into a tuple representation.

        This method is useful for serializing the object for storage or transmission.

        Returns:
            Tuple: A tuple containing the essential attributes of the Observable instance.
        """
        lst_window_end = self.interval.end
        if lst_window_end > 24:
            lst_window_end = lst_window_end - 24

        return (
            self.id,
            self.proposal_id,
            self.dt.strftime("%Y-%m-%d"),
            self.utc_constraint,
            round(self.duration, 2),
            self.observation.lst_window_start,
            self.observation.lst_window_end,
            round(self.interval.start, 2),
            round(lst_window_end, 2),
            self.interval.start_utc,
            self.interval.end_utc,
        )

    def __lt__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self.id, self.dt, self.utc_constraint, self.duration, self.interval.start) < (
            other.id,
            other.dt,
            other.utc_constraint,
            other.duration,
            other.interval.start,
        )

    def __le__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self.id, self.dt, self.utc_constraint, self.duration, self.interval.start) <= (
            other.id,
            other.dt,
            other.utc_constraint,
            other.duration,
            other.interval.start,
        )

    def __eq__(self, other: Self) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (
            self.interval.dt,
            self.utc_constraint,
            self.duration,
            self.interval.start,
            self.interval.end,
            self.observation.id,
        ) == (
            other.interval.dt,
            other.utc_constraint,
            other.duration,
            other.interval.start,
            other.interval.end,
            other.observation.id,
        )

    def __ge__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self.id, self.dt, self.utc_constraint, self.duration, self.interval.start) >= (
            other.id,
            other.dt,
            other.utc_constraint,
            other.duration,
            other.interval.start,
        )

    def __gt__(self, other) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (self.id, self.dt, self.utc_constraint, self.duration, self.interval.start) > (
            other.id,
            other.dt,
            other.utc_constraint,
            other.duration,
            other.interval.start,
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.interval.dt,
                self.utc_constraint,
                self.duration,
                self.interval.start,
                self.interval.end,
                self.observation.id,
            )
        )

    def __str__(self) -> int:
        return f"{self.interval.dt}{self.interval.type}{self.interval.start}{self.interval.end}{self.observation.id}"



"""
lstpressure.sun.Sun
"""
from typing import Optional
from .location_providers import AstralProvider, MeerKATProvider
from ..utils import normalize_coordinates, normalize_yyyymmdd_to_datetime, utc_to_lst
from perf import track_total_runtime, decorate_all
from conf import LocationProviderType as LocationProviderType, Conf

conf = Conf()

provider_mapping = {
    LocationProviderType.ASTRAL: AstralProvider,
    LocationProviderType.MEERKAT: MeerKATProvider,
}


@decorate_all(track_total_runtime)
class Sun:
    """
    Sun statistics for a particular date, at a particular lat/long.

    Attributes
    ----------
    dawn : datetime
        The dawn time for the given date and location.
    dawn_lst : datetime
        The dawn time converted to Local Sidereal Time (LST) for the given date and location.
    sunrise : datetime
        The sunrise time for the given date and location.
    sunrise_lst : datetime
        The sunrise time converted to Local Sidereal Time (LST) for the given date and location.
    noon : datetime
        The solar noon time for the given date and location.
    noon_lst : datetime
        The solar noon time converted to Local Sidereal Time (LST) for the given date and location.
    sunset : datetime
        The sunset time for the given date and location.
    sunset_lst : datetime
        The sunset time converted to Local Sidereal Time (LST) for the given date and location.
    dusk : datetime
        The dusk time for the given date and location.
    dusk_lst : datetime
        The dusk time converted to Local Sidereal Time (LST) for the given date and location.

    Parameters
    ----------
    latitude : float | str
        The latitude of the location in decimal degrees or string format.
    longitude : float | str
        The longitude of the location in decimal degrees or string format.
    yyyymmdd : str
        The date in 'YYYYMMDD' format for which sun statistics are calculated.

    Raises
    ------
    ValueError
        If the date format is incorrect or if the location coordinates cannot be normalized.

    """

    def __init__(
        self,
        yyyymmdd: str,
        latitude: Optional[float | str] = None,
        longitude: Optional[float | str] = None,
        provider: Optional[LocationProviderType] = conf.LOC_PROVIDER,
    ) -> None:
        """
        Initialize a Sun object.

        Parameters
        ----------
        yyyymmdd : str
            The date in 'YYYYMMDD' format for which sun statistics are calculated.
        latitude : Optional[float | str]
            The latitude of the location in decimal degrees or string format. Defaults to None.
        longitude : Optional[float | str]
            The longitude of the location in decimal degrees or string format. Defaults to None.
        provider: Optional[LocationProviderType]
            A sun/time provider for customizing class calculations. Defaults to LocationProviderType.ASTRAL.

        Raises
        ------
        ValueError
            If the date format is incorrect or if the location coordinates cannot be normalized.
        """
        dt = normalize_yyyymmdd_to_datetime(yyyymmdd)
        self._provider = provider_mapping.get(provider, AstralProvider)

        if latitude and longitude:
            latitude, longitude = normalize_coordinates(latitude, longitude)
        self._latitude = latitude
        self._longitude = longitude

        self._attributes = self.provider.calc_sun(
            date=dt,
            latitude=self.latitude,
            longitude=self.longitude,
        )

    def calc_intervals(self, *args, **kwargs):
        return self.provider.calc_intervals(Sun, *args, **kwargs)

    @property
    def latitude(self) -> float:
        return self._latitude

    @property
    def longitude(self) -> float:
        return self._longitude

    @property
    def provider(self) -> LocationProviderType:
        return self._provider

    def time(self, event, get_fallback=lambda: None):
        result = self._attributes.get(event, None)
        if result is None:
            result = get_fallback()
        return result

    @property
    def dawn(self):
        return self.time("dawn")

    @property
    def dawn_lst(self):
        return self.time(
            "dawn_lst", lambda: utc_to_lst(self.time("dawn"), self._latitude, self._longitude)
        )

    @property
    def sunrise(self):
        return self.time("sunrise")

    @property
    def sunrise_lst(self):
        return self.time(
            "sunrise_lst", lambda: utc_to_lst(self.time("sunrise"), self._latitude, self._longitude)
        )

    @property
    def noon(self):
        return self.time("noon")

    @property
    def noon_lst(self):
        return self.time(
            "noon_lst", lambda: utc_to_lst(self.time("noon"), self._latitude, self._longitude)
        )

    @property
    def sunset(self):
        return self.time("sunset")

    @property
    def sunset_lst(self):
        return self.time(
            "sunset_lst", lambda: utc_to_lst(self.time("sunset"), self._latitude, self._longitude)
        )

    @property
    def dusk(self):
        return self.time("dusk")

    @property
    def dusk_lst(self):
        return self.time(
            "dusk_lst", lambda: utc_to_lst(self.time("dusk"), self._latitude, self._longitude)
        )



"""
perf
"""
import datetime
import functools
import time
from conf import Conf, LogLevel
from logger import debug

module_import_time = datetime.datetime.now()
conf = Conf()

fn_runtime = {}


def monitor_perf(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if conf.LOG_LEVEL == LogLevel.DEBUG:
            debug(f"\n\n*** PERFORMANCE STATS ***")

            # Runtime
            end_time = datetime.datetime.now()
            duration = end_time - module_import_time
            debug(f"Total runtime: {duration}")

            # Find the length of the longest function identifier
            max_key_length = max(len(func_identifier) for func_identifier in fn_runtime)

            # Find the maximum length of call_count when converted to string
            max_call_count_length = max(
                len(str(call_count)) for _, (_, call_count) in fn_runtime.items()
            )

            # Sort items by total runtime in descending order
            sorted_items = sorted(fn_runtime.items(), key=lambda item: item[1][0], reverse=True)

            # Print each entry with padding for alignment
            for func_identifier, (total_runtime, call_count) in sorted_items:
                padded_func_identifier = func_identifier.ljust(max_key_length + 1)
                padded_call_count = str(call_count).rjust(max_call_count_length)
                debug(
                    f" :: {padded_func_identifier} {total_runtime:.4f} seconds : {padded_call_count} calls"
                )
        return result

    return wrapper


def track_total_runtime(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()

        if conf.LOG_LEVEL == LogLevel.DEBUG:
            # Construct a unique key using module name and function name
            unique_key = f"{func.__module__}.{func.__name__}"

            # Update cumulative runtime and call count in the global dictionary
            runtime = end_time - start_time
            if unique_key in fn_runtime:
                fn_runtime[unique_key] = (
                    fn_runtime[unique_key][0] + runtime,
                    fn_runtime[unique_key][1] + 1,
                )
            else:
                fn_runtime[unique_key] = (runtime, 1)
        return result

    return wrapper


def decorate_all(decorator):
    def decorate(cls):
        for attr_name, attr_value in cls.__dict__.items():
            if isinstance(attr_value, staticmethod):
                # Unwrap the static method, decorate it, and then re-wrap it
                original_func = attr_value.__func__
                decorated_func = decorator(original_func)
                setattr(cls, attr_name, staticmethod(decorated_func))
            elif isinstance(attr_value, classmethod):
                # Unwrap the class method, decorate it, and then re-wrap it
                original_func = attr_value.__func__
                decorated_func = decorator(original_func)
                setattr(cls, attr_name, classmethod(decorated_func))
            elif isinstance(attr_value, property):
                # Apply the decorator to the getter, and re-create the property
                original_get = attr_value.fget
                decorated_get = decorator(original_get) if original_get else None
                original_set = attr_value.fset
                decorated_set = decorator(original_set) if original_set else None
                original_del = attr_value.fdel
                decorated_del = decorator(original_del) if original_del else None
                setattr(
                    cls,
                    attr_name,
                    property(decorated_get, decorated_set, decorated_del, attr_value.__doc__),
                )
            elif callable(attr_value):
                # Regular method
                setattr(cls, attr_name, decorator(attr_value))
        return cls

    return decorate


__all__ = ["monitor_perf", "track_total_runtime", "decorate_all"]

# Automatically added by katversion
__version__ = '1.3.0'



"""
lstpressure.lstindex.LSTIntervalType
"""

from enum import Enum, auto


class LSTIntervalType(Enum):
    """
    Enumeration of LST (Local Sidereal Time) interval types.

    Defines various types of LST intervals that can be associated with observations or calendar events.

    Attributes
    ----------
    ALL_DAY : LSTIntervalType
        An interval type representing a full day, spanning from 0 to 24 hours.
    NIGHT : LSTIntervalType
        An interval type representing the nighttime period.
    OBSERVATION_WINDOW : LSTIntervalType
        An interval type representing an observation window.
    SUNRISE_SUNSET : LSTIntervalType
        An interval type representing the period from sunrise to sunset.
    SUNSET_SUNRISE : LSTIntervalType
        An interval type representing the period from sunset to sunrise.
    """

    ALL_DAY = auto()
    NIGHT = auto()
    OBSERVATION_WINDOW = auto()
    SUNRISE_SUNSET = auto()
    SUNSET_SUNRISE = auto()



from lstpressure import lstcalendar, lst, lstindex, observable, observation, utils, sun
from lstpressure.lstcalendar import LSTCalendar
from lstpressure.lst import LST
from lstpressure.lstindex import LSTIntervalType
from lstpressure.observable.Observable import Observable
from lstpressure.observation import is_observable, Observation
from lstpressure.sun import Sun
from conf import LocationProviderType, Conf as LSTConf

try:
    import katversion as _katversion
except ImportError:
    import time as _time

    __version__ = "0.0+unknown.{}".format(_time.strftime("%Y%m%d%H%M"))
else:
    __version__ = _katversion.get_version()

__all__ = [
    "__version__",
    "is_observable",
    "lstcalendar",
    "LSTCalendar",
    "LST",
    "lst",
    "lstindex",
    "LSTIntervalType",
    "observable",
    "Observable",
    "observables",
    "observation",
    "Observation",
    "sun",
    "Sun",
    "utils",
    "LSTConf",
    "LocationProviderType",
]

# Automatically added by katversion
__version__ = '1.3.0'



import pytest
from lstpressure.sun import Sun
from lstpressure.sun.location_providers import MeerKATProvider
from lstpressure.sun.location_providers.normalize_intervals import normalize_interval
from lstpressure.sun.location_providers.astral_provider.AstralProvider import (
    normalize_interval,
    AstralProvider,
)
from conf import Conf, LocationProviderType
from lstpressure.utils import normalize_yyyymmdd_to_datetime

conf = Conf()
conf.LOC_PROVIDER = LocationProviderType.ASTRAL

latitude, longitude = ["-30:42:39.8", "21:26:38.0"]

# NOTE - all times are in UTC unless suffixed with "_lst"

data = [
    (
        "20231204",
        {
            "meerkat_provider": {
                "dawn": None,
                "sunrise": "2023-12-04 03:27:00.000000",
                "sunrise_lst": 9.74,
                "noon": None,
                "sunset": "2023-12-04 17:17:00.000000",
                "sunset_lst": 23.6,
                "dusk": None,
            },
        },
    ),
    (
        "20231121",
        {
            "astral_provider": {
                "dawn": "2023-11-21 02:59:16.055392",
                "sunrise": "2023-11-21 03:26:25.080197",
                "noon": "2023-11-21 10:19:52.000000",
                "sunset": "2023-11-21 17:13:54.229446",
                "dusk": "2023-11-21 17:41:06.904472",
            },
            "meerkat_provider": {
                "dawn": None,
                "sunrise": "2023-11-21 03:29:00.000000",
                "noon": None,
                "sunset": "2023-11-21 17:06:00.000000",
                "dusk": None,
            },
        },
    ),
    (
        "20230615",
        {
            "astral_provider": {
                "dawn": "2023-06-15 05:02:41.873453",
                "sunrise": "2023-06-15 05:29:46.909659",
                "noon": "2023-06-15 10:34:34.000000",
                "sunset": "2023-06-15 15:39:30.356171",
                "dusk": "2023-06-15 16:06:35.427817",
            },
            "meerkat_provider": {
                "dawn": None,
                "sunrise": "2023-06-15 05:32:00.000000",
                "noon": None,
                "sunset": "2023-06-15 15:32:00.000000",
                "dusk": None,
            },
        },
    ),
]


@pytest.mark.parametrize(
    "yyyymmdd, expected_results",
    data,
)
def test_Sun_with_AstralProvider(yyyymmdd, expected_results):
    if expected_results.get('"astral_provider"'):
        sun = Sun(yyyymmdd, latitude, longitude)
        for event, expected_time in expected_results["astral_provider"].items():
            calculated_time = getattr(sun, event).strftime("%Y-%m-%d %H:%M:%S.%f")
            assert calculated_time == expected_time


@pytest.mark.parametrize(
    "yyyymmdd, expected_results",
    data,
)
def test_Sun_with_MeerKATProvider(yyyymmdd, expected_results):
    sun = Sun(yyyymmdd=yyyymmdd, provider=LocationProviderType.MEERKAT)

    for event, expected_time in expected_results["meerkat_provider"].items():
        try:
            calculated_time = getattr(sun, event)
            if isinstance(calculated_time, float):
                assert expected_time == round(calculated_time, 2)
            else:
                assert expected_time == calculated_time.strftime("%Y-%m-%d %H:%M:%S.%f")

        except AttributeError:
            calculated_time = None


@pytest.mark.parametrize(
    "start, end, expected",
    [
        (7, 13, (7, 13)),
        (22, 7, (22, 31)),
        (1, 6, (1, 6)),
        (18, 2, (18, 26)),
    ],
)
def test_normalize_interval(start, end, expected):
    result = normalize_interval(start, end)
    assert result == expected


@pytest.mark.parametrize(
    "latitude,longitude,today_dt",
    [
        ("-30:42:39.8", "21:26:38.0", "20231116"),
    ],
)
def test_calculate_intervals(latitude, longitude, today_dt):
    today = normalize_yyyymmdd_to_datetime(today_dt)
    intervals = AstralProvider.calc_intervals(Sun, latitude, longitude, today, None)
    # TODO - what to assert



import pytest
from datetime import datetime, date
from lstpressure.utils import normalize_datetime, normalize_yyyymmdd_to_datetime


@pytest.mark.parametrize(
    "input,expected", [("2023-12-20 00:00:00", datetime(2023, 12, 20, 0, 0, 0))]
)
def test_normalize_datetime(input, expected):
    result = normalize_datetime(input)
    assert result == expected


@pytest.mark.parametrize("input,expected", [("20230222", date(2023, 2, 22))])
def test_normalize_yyyymmdd_to_datetime(input, expected):
    result = normalize_yyyymmdd_to_datetime(input)
    assert result == expected



"""
lstpressure.lstindex.LSTIndex
------

Contains the LSTIndex class, which acts as a wrapper around the IntervalTree class for simplified interval operations.
"""

from intervaltree import IntervalTree, Interval
from perf import track_total_runtime, decorate_all


@decorate_all(track_total_runtime)
class LSTIndex:
    """
    A wrapper around the IntervalTree class for simplified interval operations.

    This class provides convenience methods to insert and query intervals.
    It internally uses the IntervalTree for interval management and query operations.

    Attributes
    ----------
    idx : IntervalTree
        The internal interval tree used for managing intervals.
    tree : IntervalTree
        Alias for the idx attribute.

    Methods
    -------
    get_entries() -> set
        Retrieve all intervals stored in the interval tree.

        Returns
        -------
        set
            A set of (Interval, data) pairs.
    insert(*args)
        Insert an interval into the interval tree.

        This method supports multiple calling patterns:
        - With a single Interval argument.
        - With two arguments specifying the start and end of the interval.
        - With three arguments specifying the start, end, and data associated with the interval.

        Parameters
        ----------
        *args
            Variable length argument list.

        Raises
        ------
        ValueError
            If the arguments don't match any of the accepted input patterns.
    envelop(*args) -> set
        Retrieve intervals that are enveloped by the given interval.

        Parameters
        ----------
        *args
            Variable length argument list.

        Returns
        -------
        set
            A set of (Interval, data) pairs.
    overlap(*args) -> set
        Retrieve intervals that overlap with the given interval.

        Parameters
        ----------
        *args
            Variable length argument list.

        Returns
        -------
        set
            A set of (Interval, data) pairs.
    """

    def __init__(self):
        """
        Initializes an empty interval tree.
        """
        self.idx = IntervalTree()
        self.tree = self.idx

    def get_entries(self) -> set:
        """
        Retrieve all intervals stored in the interval tree.

        Returns
        -------
        set
            A set of (Interval, data) pairs.
        """
        return self.idx.items()

    def insert(self, *args):
        """
        Insert an interval into the interval tree.

        This method supports multiple calling patterns:
        - With a single Interval argument.
        - With two arguments specifying the start and end of the interval.
        - With three arguments specifying the start, end, and data associated with the interval.

        Parameters
        ----------
        *args
            Variable length argument list.

        Raises
        ------
        ValueError
            If the arguments don't match any of the accepted input patterns.
        """
        if len(args) == 1 and isinstance(args[0], Interval):
            interval = args[0]
            self.idx.add(interval)
        elif len(args) == 2:
            begin, end = args
            self.idx.addi(begin, end, {})
        elif len(args) == 3:
            begin, end, data = args
            self.idx.addi(begin, end, data)
        else:
            raise ValueError("Invalid arguments")

    def envelop(self, *args) -> set:
        """
        Retrieve intervals that are enveloped by the given interval.

        Parameters
        ----------
        *args
            Variable length argument list.

        Returns
        -------
        set
            A set of (Interval, data) pairs.
        """
        return self.idx.envelop(*args)

    def overlap(self, *args) -> set:
        """
        Retrieve intervals that overlap with the given interval.

        Parameters
        ----------
        *args
            Variable length argument list.

        Returns
        -------
        set
            A set of (Interval, data) pairs.
        """
        return self.idx.overlap(*args)



"""
lstpressure.lstcalendar.LSTCalendarDate
"""
from datetime import timedelta, date
from lstpressure.lstindex import LSTInterval
from ..lstcalendar import LSTCalendar
from ..observable import Observable
from ..sun import Sun
from perf import decorate_all, track_total_runtime
from conf import Conf, LogLevel, LocationProviderType
from logger import info
from typing import Optional

conf = Conf()


@decorate_all(track_total_runtime)
class LSTCalendarDate:
    c = 0
    """
    Represents a date within the LSTCalendar.

    Parameters
    ----------
    dt : date
        The date for which LSTCalendarDate is instantiated.
    cal : LSTCalendar
        The parent LSTCalendar object that manages this date.

    Attributes
    ----------
    dt : date
        The date for which LSTCalendarDate is instantiated.
    tomorrow_dt : date
        The date following the current date.
    sun : Sun
        An object representing sun statistics for the current date.
    tomorrow_sun : Sun
        An object representing sun statistics for the following date.
    calendar : LSTCalendar
        The parent LSTCalendar object that manages this date.
    intervals : list[LSTInterval]
        A list of LSTInterval objects associated with this date.

    Methods
    -------
    observables() -> list[Observable]
        Retrieve observable observations for this date.
    to_yyyymmdd() -> str
        Convert the date to a string in the 'YYYYMMDD' format.
    """

    def __init__(self, dt, cal, provider: Optional[LocationProviderType] = None) -> None:
        """
        Initialize an LSTCalendarDate object.

        Parameters
        ----------
        dt : date
            The date for which LSTCalendarDate is instantiated.
        cal : LSTCalendar
            The parent LSTCalendar object that manages this date.
        provider : Optional[LocationProviderType]. Defaults to None

        Raises
        ------
        TypeError
            If the "cal" argument is missing.
        """
        LSTCalendarDate.c += 1
        if not cal:
            raise TypeError(
                'Missing "cal" argument, LSTCalendarDate instances must be instantiated via instances of LSTCalendar so that the self.cal can be assigned'
            )

        provider = provider if provider else conf.LOC_PROVIDER
        self.dt: date = dt
        self.tomorrow_dt = dt + timedelta(days=1)
        self.sun = Sun(dt, cal.latitude, cal.longitude, provider=provider)
        self.tomorrow_sun = Sun(
            dt + timedelta(days=1), cal.latitude, cal.longitude, provider=provider
        )
        self.calendar: LSTCalendar = cal
        self.intervals: list[LSTInterval] = self.sun.calc_intervals(
            cal.latitude, cal.longitude, dt, self
        )
        if conf.LOG_LEVEL != LogLevel.WARN:
            if LSTCalendarDate.c % 25 == 0:
                info(f" ... {dt}")
        for interval in self.intervals:
            cal.interval_index.insert(interval.interval)

    def observables(self) -> list[Observable]:
        """
        Retrieve observable observations for this date.

        Returns
        -------
        list[Observable]
            A list of observable observations for this date.
        """
        result = set()

        # Cycle through dt intervals
        for cal_interval in self.intervals:
            cal_interval_end = cal_interval.end
            interval_type = cal_interval.type
            query = cal_interval.interval

            # Note that overlap() returns a Set
            query_result = self.calendar.observations_index.overlap(query)

            for obs_interval_raw in query_result:
                (
                    obs_window_start,
                    obs_window_end,
                    obs_interval,
                ) = obs_interval_raw
                obs = obs_interval.parent
                utc_constraints = obs.utc_constraints
                duration = obs.duration

                if (utc_constraints is None or len(utc_constraints) == 0) or (
                    len(utc_constraints) > 0 and interval_type in utc_constraints
                ):
                    if query.end > obs_window_start:
                        if (
                            obs_window_start + duration < cal_interval_end
                            or obs_window_end + duration < cal_interval_end
                        ):
                            result.add(Observable(cal_interval, obs))

        return list(result)

    def to_yyyymmdd(self) -> str:
        """
        Convert the date to a string in the 'YYYYMMDD' format.

        Returns
        -------
        str
            The date in 'YYYYMMDD' format.
        """
        return self.dt.strftime("%Y%m%d")



from ....lstindex import LSTInterval
from datetime import datetime, timedelta
from ..LocationProvider import LocationProvider
from astral.sun import sun as calc_sun
from astral import LocationInfo
from perf import track_total_runtime, decorate_all
from ....utils import normalize_coordinates
from ..normalize_intervals import normalize_interval


@decorate_all(track_total_runtime)
class AstralProvider(LocationProvider):
    @staticmethod
    def calc_sun(latitude, longitude, date):
        location = LocationInfo(latitude=latitude, longitude=longitude)
        location.timezone = "UTC"
        return calc_sun(location.observer, date=date)

    @staticmethod
    def calc_intervals(
        Sun, latitude: str | float, longitude: str | float, today_dt: datetime, obj
    ) -> list[LSTInterval]:
        # Interpreter sees this as a module if at the top of the file
        from ....lstindex import LSTIntervalType, LSTInterval

        latitude, longitude = normalize_coordinates(latitude, longitude)
        today = today_dt
        today_sun = Sun(today, latitude, longitude, provider=AstralProvider)
        today_sunrise_lst = today_sun.sunrise_lst
        today_sunrise_utc = today_sun.sunrise
        today_sunset_lst = today_sun.sunset_lst
        today_sunset_utc = today_sun.sunset
        today_dusk_lst = today_sun.dusk_lst
        today_dusk_utc = today_sun.dusk

        tomorrow = today + timedelta(days=1)
        tomorrow_sun = Sun(tomorrow, latitude, longitude, provider=AstralProvider)
        tomorrow_dawn_lst = tomorrow_sun.dawn_lst
        tomorrow_dawn_utc = tomorrow_sun.dawn
        tomorrow_sunrise_lst = tomorrow_sun.sunrise_lst
        tomorrow_sunrise_utc = tomorrow_sun.sunrise
        tomorrow_sunset_lst = tomorrow_sun.sunset_lst
        tomorrow_sunset_utc = tomorrow_sun.sunset

        result = []

        # ALL DAY
        result.append(
            LSTInterval(
                0, 24, None, None, obj, today, LSTIntervalType.ALL_DAY, today_sun, tomorrow_sun
            )
        )

        # SUNRISE_SUNSET
        SUNRISE_SUNSET = LSTInterval(
            *normalize_interval(today_sunrise_lst, today_sunset_lst),
            today_sunrise_utc.strftime("%H:%M"),
            today_sunset_utc.strftime("%H:%M"),
            obj,
            today,
            LSTIntervalType.SUNRISE_SUNSET,
            today_sun,
            tomorrow_sun,
        )
        result.append(SUNRISE_SUNSET)
        if SUNRISE_SUNSET.end > 24:
            result.append(
                LSTInterval(
                    0,
                    tomorrow_sunset_lst,
                    today_sunrise_utc.strftime("%H:%M"),
                    tomorrow_sunset_utc.strftime("%H:%M"),
                    obj,
                    today,
                    LSTIntervalType.SUNRISE_SUNSET,
                    today_sun,
                    tomorrow_sun,
                )
            )

        # SUNSET_SUNRISE
        SUNSET_SUNRISE = LSTInterval(
            *normalize_interval(today_sunset_lst, tomorrow_sunrise_lst),
            today_sunset_utc.strftime("%H:%M"),
            tomorrow_sunrise_utc.strftime("%H:%M"),
            obj,
            today,
            LSTIntervalType.SUNSET_SUNRISE,
            today_sun,
            tomorrow_sun,
        )
        result.append(SUNSET_SUNRISE)

        if SUNSET_SUNRISE.end > 24:
            result.append(
                LSTInterval(
                    0,
                    tomorrow_sunrise_lst,
                    0,
                    tomorrow_sunrise_utc.strftime("%H:%M"),
                    obj,
                    today,
                    LSTIntervalType.SUNSET_SUNRISE,
                    today_sun,
                    tomorrow_sun,
                )
            )

        # NIGHT
        NIGHT = LSTInterval(
            *normalize_interval(today_dusk_lst, tomorrow_dawn_lst),
            today_dusk_utc.strftime("%H:%M"),
            tomorrow_dawn_utc.strftime("%H:%M"),
            obj,
            today,
            LSTIntervalType.NIGHT,
            today_sun,
            tomorrow_sun,
        )
        result.append(NIGHT)

        if NIGHT.end > 24:
            result.append(
                LSTInterval(
                    0,
                    tomorrow_dawn_lst,
                    0,
                    tomorrow_dawn_utc.strftime("%H:%M"),
                    obj,
                    today,
                    LSTIntervalType.NIGHT,
                    today_sun,
                    tomorrow_sun,
                )
            )

        return result



from ..LocationProvider import LocationProvider
from datetime import date, datetime, time, timedelta
from .DateTimeUtils import DateTimeUtil
from ....lstindex import LSTInterval
from .SolarSystemUtils import SolarSystem
from .AstroUtils import AstroUtil
import json
import os
from logger import warn
from perf import track_total_runtime, decorate_all
from ..normalize_intervals import normalize_interval
from conf import LocationProviderType as S


@decorate_all(track_total_runtime)
class MeerKATProvider(LocationProvider):
    @staticmethod
    def calc_sun(date: date, latitude, longitude):
        date = datetime.combine(date, time(0))
        with open(os.path.join(os.path.dirname(__file__), "planets.json"), "r") as f:
            planets = json.load(f)

            jd = DateTimeUtil.jd(date)
            sun_pos = SolarSystem.sunpos(planets["Earth"], jd)
            sun_lst = calc_sun_rise_set(
                sun_pos["ra"], sun_pos["dec"], AstroUtil.MKAT_POSITION["latitude"], 0
            )

            set_utc = DateTimeUtil.lst2ut(
                sun_lst["lstSet"], AstroUtil.MKAT_POSITION["longitude"], date
            )
            rise_utc = DateTimeUtil.lst2ut(
                sun_lst["lstRise"], AstroUtil.MKAT_POSITION["longitude"], date
            )

            set_date = date.replace(hour=int(set_utc), minute=int((set_utc - int(set_utc)) * 60))
            rise_date = date.replace(
                hour=int(rise_utc), minute=int((rise_utc - int(rise_utc)) * 60)
            )

            return {
                "dawn": None,
                "dusk": None,
                "noon": None,
                "sunrise": rise_date,
                "sunrise_lst": sun_lst["lstRise"],
                "sunset": set_date,
                "sunset_lst": sun_lst["lstSet"],
            }

    @staticmethod
    def calc_intervals(
        Sun, latitude: str | float, longitude: str | float, today_dt: datetime, obj=None
    ) -> list[LSTInterval]:
        # Interpreter sees this as a module if at the top of the file
        from ....lstindex import LSTIntervalType, LSTInterval

        if latitude or longitude:
            warn(
                "MEERKAT sub provider doesn't allow for overriding latitude and longitude values. Provided values will be ignored."
            )

        today = today_dt
        today_sun = Sun(today, provider=S.MEERKAT)
        today_sunrise_lst = today_sun.sunrise_lst
        today_sunrise_utc = today_sun.sunrise
        today_sunset_lst = today_sun.sunset_lst
        today_sunset_utc = today_sun.sunset

        tomorrow = today + timedelta(days=1)
        tomorrow_sun = Sun(tomorrow, provider=S.MEERKAT)
        tomorrow_sunrise_lst = tomorrow_sun.sunrise_lst
        tomorrow_sunrise_utc = tomorrow_sun.sunrise
        tomorrow_sunset_lst = tomorrow_sun.sunset_lst
        tomorrow_sunset_utc = tomorrow_sun.sunset

        result = []

        # ALL DAY
        result.append(
            LSTInterval(
                0, 24, None, None, obj, today, LSTIntervalType.ALL_DAY, today_sun, tomorrow_sun
            )
        )

        # SUNRISE_SUNSET
        SUNRISE_SUNSET = LSTInterval(
            *normalize_interval(today_sunrise_lst, today_sunset_lst),
            today_sunrise_utc.strftime("%H:%M"),
            today_sunset_utc.strftime("%H:%M"),
            obj,
            today,
            LSTIntervalType.SUNRISE_SUNSET,
            today_sun,
            tomorrow_sun,
        )
        result.append(SUNRISE_SUNSET)
        if SUNRISE_SUNSET.end > 24:
            result.append(
                LSTInterval(
                    0,
                    tomorrow_sunset_lst,
                    today_sunrise_utc.strftime("%H:%M"),
                    tomorrow_sunset_utc.strftime("%H:%M"),
                    obj,
                    today,
                    LSTIntervalType.SUNRISE_SUNSET,
                    today_sun,
                    tomorrow_sun,
                )
            )

        # SUNSET_SUNRISE
        SUNSET_SUNRISE = LSTInterval(
            *normalize_interval(today_sunset_lst, tomorrow_sunrise_lst),
            today_sunset_utc.strftime("%H:%M"),
            tomorrow_sunrise_utc.strftime("%H:%M"),
            obj,
            today,
            LSTIntervalType.SUNSET_SUNRISE,
            today_sun,
            tomorrow_sun,
        )
        result.append(SUNSET_SUNRISE)

        if SUNSET_SUNRISE.end > 24:
            result.append(
                LSTInterval(
                    0,
                    tomorrow_sunrise_lst,
                    0,
                    tomorrow_sunrise_utc.strftime("%H:%M"),
                    obj,
                    today,
                    LSTIntervalType.SUNSET_SUNRISE,
                    today_sun,
                    tomorrow_sun,
                )
            )

        # NIGHT
        NIGHT = LSTInterval(
            *normalize_interval(today_sunset_lst, tomorrow_sunrise_lst),
            today_sunset_utc.strftime("%H:%M"),
            tomorrow_sunrise_utc.strftime("%H:%M"),
            obj,
            today,
            LSTIntervalType.NIGHT,
            today_sun,
            tomorrow_sun,
        )
        result.append(NIGHT)

        if NIGHT.end > 24:
            result.append(
                LSTInterval(
                    0,
                    tomorrow_sunrise_lst,
                    0,
                    tomorrow_sunrise_utc.strftime("%H:%M"),
                    obj,
                    today,
                    LSTIntervalType.NIGHT,
                    today_sun,
                    tomorrow_sun,
                )
            )

        return result


@track_total_runtime
def calc_sun_rise_set(
    ra: float | str,
    dec: float | str,
    latitude: float | str,
    thresh_hold: float | str,
):
    cos_h = -(AstroUtil.sind(thresh_hold) + AstroUtil.sind(latitude) * AstroUtil.sind(dec)) / (
        AstroUtil.cosd(latitude) * AstroUtil.cosd(dec)
    )
    obj = {
        "never_up": False,
        "circumpolar": False,
        "lstRise": 0,
        "lstSet": 0,
    }
    ha = AstroUtil.acosd(cos_h) / 15.0
    ra_hours = ra / 15.0  # Convert RA to hours
    obj["lstRise"] = AstroUtil.rev24(ra_hours - ha)
    obj["lstSet"] = AstroUtil.rev24(ra_hours + ha)
    return obj



from .meerkat_provider.MeerKATProvider import MeerKATProvider
from .astral_provider.AstralProvider import AstralProvider
from .LocationProvider import LocationProvider

__all__ = ["MeerKATProvider", "LocationProvider", "AstralProvider"]

# Automatically added by katversion
__version__ = '1.3.0'



from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))


def get_long_description():
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()
        return long_description


setup(
    name="lst-pressure",
    setup_requires=["katversion"],
    use_katversion=True,
    description='Determine periods of "LST pressure" by querying for intersections between LST/Solar intervals',
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/ska-sa/lst-pressure",
    author="Zach Smith",
    author_email="zsmith@sarao.ac.za",
    license="Apache 2.0",
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.12",
    ],
    package_dir={},
    packages=find_packages(where="./"),
    python_requires=">=3.12",
)



"""
utils.normalize_coordinates
"""
from functools import lru_cache
from perf import track_total_runtime


@track_total_runtime
@lru_cache(maxsize=None)
def normalize_coordinates(lat, long) -> tuple[float, float]:
    """fn"""

    def is_decimal(value):
        """Check if a value is already in decimal format"""
        try:
            float(value)
            return True
        except ValueError:
            return False

    def dms_str_to_dec(dms_str):
        """Convert DMS string to decimal value"""
        negative = dms_str.startswith("-")
        dms_str = dms_str.lstrip("-")
        degrees, minutes, seconds = map(float, dms_str.split(":"))
        decimal = degrees + (minutes / 60) + (seconds / 3600)
        return -decimal if negative else decimal

    if is_decimal(lat) or is_decimal(long):
        if not is_decimal(lat) and is_decimal(long):
            raise ValueError(
                "lat and long must both be in DMS or decimal format, not a mix of the two."
            )
        return (float(lat), float(long))

    return (dms_str_to_dec(lat), dms_str_to_dec(long))



import argparse
import logging
from conf import Conf, LogLevel
from .apps import __all__ as app_lib
from lstpressure import __version__ as version

conf = Conf()


class CustomHelpFormatter(argparse.HelpFormatter):
    def _format_action(self, action):
        if isinstance(action, argparse._SubParsersAction):
            parts = []
            for subaction in action._choices_actions:
                parts.append(self._format_action(subaction))
            return "".join(parts)
        else:
            return super()._format_action(action)


def configure_logging(log_level):
    numeric_level = getattr(logging, log_level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f"Invalid log level: {log_level}")
    logging.basicConfig(level=numeric_level)


def cli():
    prog = "lstpressure"
    description = f"(NRF SARAO) LST Pressure v{version}"

    parser = argparse.ArgumentParser(
        prog=prog, description=description, formatter_class=CustomHelpFormatter
    )
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("--debug", action="store_true", help='Show logs of "DEBUG" level or higher')
    parser.add_argument(
        "--info", action="store_true", help='Show logs of "INFO" level or higher (default)'
    )
    parser.add_argument("--warn", action="store_true", help='Show logs of "WARN" level or higher')

    module_parser = parser.add_subparsers(title="Modules", dest="command")

    # Build the CLI
    apps = {
        app.id[0:3]: app(
            module_parser.add_parser(
                app.id,
                usage=app.usage,
                description=app.description,
                help=app.description,
                aliases=[app.id[0:3]],
            )
        )
        for app in app_lib
    }
    args = parser.parse_args()

    # Override environment configuration
    if args.warn:
        conf.LOG_LEVEL = LogLevel.WARN
    if args.info:
        conf.LOG_LEVEL = LogLevel.INFO
    if args.debug:
        conf.LOG_LEVEL = LogLevel.DEBUG

    # Try get the application
    try:
        app = apps.get(args.command[0:3], None)
    except:
        app = None

    # If app not specified correctly then print help and exit
    if not app:
        parser.print_help()
        return

    # Otherwise execute the application
    app.parse(args).exe()

# Automatically added by katversion
__version__ = '1.3.0'



import pytest
from lstpressure.lstcalendar import LSTCalendar
from lstpressure.observation import Observation, is_observable
from lstpressure.lstindex import LSTIntervalType as I
from conf import LocationProviderType

# 20231030 LST dusk is about 2130

tests = [
    # (lst_window_start, lst_window_end, utc_constraints, duration, dt_range, observables_count)
    (8, 20, [I.NIGHT], 2, ["20231030"], 0),
    (8, 20, [], 2, ["20231030"], 2),
    (2, 20, [I.NIGHT], 0.5, ["20231030"], 1),
    (20, 1, [I.NIGHT], 2, ["20231030"], 2),
    (20, 1, [I.NIGHT], 2, ["20231030", "20231031"], 4),
    (20, 1, [], 2, ["20231030", "20231031"], 10),
    (20, 1, [], 2, ["20231106"], 5),
    (20, 1, [I.SUNRISE_SUNSET], 2, ["20231030", "20231031"], 0),
    (20, 1, [I.SUNSET_SUNRISE], 2, ["20231030", "20231031"], 4),
    (12.5, 15.5, None, None, ["20231107"], 2),  # FROM OPT
    (
        3.75,
        5.25,
        [I.NIGHT],
        5.655,
        ["20240501", "20240701"],
        0,
    ),  # OPT (https://github.com/ska-sa/lst-pressure/issues/61)
]


@pytest.mark.parametrize(
    "lst_window_start, lst_window_end, utc_constraints, duration, dt_range, observables_count",
    tests,
)
def test_is_observable(
    lst_window_start,
    lst_window_end,
    utc_constraints,
    duration,
    dt_range,
    observables_count,
):
    assert is_observable(
        Observation("~", lst_window_start, lst_window_end, utc_constraints, duration),
        *dt_range,
        provider=LocationProviderType.ASTRAL,
    ) is bool(observables_count)


@pytest.mark.parametrize(
    "lst_window_start, lst_window_end, utc_constraints, duration, dt_range, observables_count",
    tests,
)
def test_observation_observables(
    lst_window_start,
    lst_window_end,
    utc_constraints,
    duration,
    dt_range,
    observables_count,
):
    assert (
        len(
            sorted(
                Observation(
                    "~", lst_window_start, lst_window_end, utc_constraints, duration
                ).observables(
                    lstcalendar=LSTCalendar(*dt_range, provider=LocationProviderType.ASTRAL)
                )
            )
        )
        == observables_count
    )


@pytest.mark.parametrize(
    "lst_window_start, lst_window_end, utc_constraints, duration, dt_range, observables_count",
    tests,
)
def test_calendar_observables(
    lst_window_start,
    lst_window_end,
    utc_constraints,
    duration,
    dt_range,
    observables_count,
):
    assert (
        len(
            sorted(
                LSTCalendar(*dt_range, provider=LocationProviderType.ASTRAL).observables(
                    [Observation("~", lst_window_start, lst_window_end, utc_constraints, duration)]
                )
            )
        )
        == observables_count
    )
